import argparse
import os

from slack_transfer import DownloaderClient


def set_parser_emoji(parser: argparse.ArgumentParser) -> None:
    """See emoji section in :doc:`../../../user_guide/cli`"""
    parser.add_argument(
        "--data_dir",
        "-d",
        type=str,
        required=True,
        help="Data directory for saving download data or loading upload data. This is required.",
    )
    parser.add_argument(
        "--downloader_token",
        "-td",
        type=str,
        required=True,
        help="Download token obtained from slack (the original workspace). Starts with `xoxb-`",
    )


def main_emoji(args: argparse.Namespace) -> None:
    """See emoji section in :doc:`../../../user_guide/cli`"""

    local_data_dir = args.data_dir
    os.makedirs(local_data_dir, exist_ok=True)
    downloader = DownloaderClient(
        local_data_dir=local_data_dir, token=args.downloader_token
    )
    downloader.download_emoji()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    set_parser_emoji(parser=parser)
    args = parser.parse_args()
    main_emoji(args=args)



import json
import os.path
import ssl
import tkinter
from tkinter import filedialog
from typing import Dict
from typing import List
from typing import Tuple
from typing import TypeVar

import certifi
from prompt_toolkit import cursor_shapes
from prompt_toolkit.shortcuts import button_dialog
from prompt_toolkit.shortcuts import checkboxlist_dialog
from prompt_toolkit.shortcuts import input_dialog
from prompt_toolkit.shortcuts import message_dialog
from prompt_toolkit.shortcuts import radiolist_dialog
from prompt_toolkit.styles import Style

from slack_transfer import DownloaderClient
from slack_transfer._base import CommonNoLocalVolumeClient
from slack_transfer.run import run
from slack_transfer.version import __version__

T = TypeVar("T")
style = Style.from_dict({"dialog.body": "bg:#cccccc #000000"})


def radio_selection(
    title: str, text: str, choices_and_values: List[Tuple[str, T]]
) -> T:
    result: T = radiolist_dialog(
        title=title,
        text=text,
        values=[(y, x) for x, y in choices_and_values],
        style=style,
    ).run()
    if result is None:
        raise KeyboardInterrupt
    return result


def multiple_selection(
    title: str, text: str, choices_and_values: List[Tuple[str, T]]
) -> List[T]:

    dialog = checkboxlist_dialog(
        title=title,
        text=text,
        values=[(y, x) for x, y in choices_and_values],
        style=style,
    )
    dialog.cursor = cursor_shapes.SimpleCursorShapeConfig(
        cursor_shape=cursor_shapes.CursorShape.BLINKING_BLOCK
    )
    result: List[T] = dialog.run()
    if result is None:
        raise KeyboardInterrupt
    return result


def button_selection(
    title: str, text: str, choices_and_values: List[Tuple[str, T]]
) -> T:
    result: T = button_dialog(
        title=title, text=text, buttons=choices_and_values, style=style
    ).run()
    return result


def confirmation(title: str, text: str) -> None:
    message_dialog(title=title, text=text, style=style).run()
    return


def interactive() -> None:
    """See interactive section in :doc:`/user_guide/cli`"""
    try:
        ssl_context = ssl.create_default_context(
            ssl.Purpose.SERVER_AUTH, cafile=certifi.where(), capath=certifi.where()
        )
        confirmation(
            title="Start",
            text="In this system, all selection can be done using mouse pointer.\nこのシステムではすべての選択をマウスで行うことができます．",
        )

        language = button_selection(
            title="Select language",
            text="Please select language",
            choices_and_values=[("English", "en"), ("日本語", "ja")],
        )

        confirmation(
            title={"en": "Select folder", "ja": "フォルダー選択"}[language],
            text={
                "en": "Next, please select your folder to save and load the downloaded data. "
                + "We highly recommend making and selecting an empty folder when this is the first time.",
                "ja": "次に，このシステムで使用するフォルダーを選択してください．初めての使用であれば，空のフォルダーを作成し，選択することを強くお勧めします．",
            }[language],
        )

        tk = tkinter.Tk()
        local_data_dir = filedialog.askdirectory()
        tk.destroy()

        already_downloaded = os.path.exists(
            os.path.join(local_data_dir, "channels.json")
        ) and os.path.exists(os.path.join(local_data_dir, "members.json"))

        if already_downloaded:
            skip_download = button_selection(
                title="Download Selection",
                text={
                    "en": "Do you want to download slack data? Usually select yes. "
                    + "When you already downloaded data, you can select No.",
                    "ja": "Slackデータのダウンロードを行いますか？通常はYesを選択してください．すでにデータをダウンロード済みの場合はNoを選択しても構いません．",
                }[language],
                choices_and_values=[("Yes", False), ("No", True)],
            )
        else:
            skip_download = False

        downloader_token = None
        if not skip_download:
            downloader_token = input_dialog(
                title="Input download token",
                text={
                    "en": "Please inputs download token. How to get is: \n"
                    + "https://slack-transfer.readthedocs.io/en/stable/user_guide/run.html#downloader-token",
                    "ja": "ダウンロード用のAPI tokenを入力してください．\n詳細: "
                    + "https://slack-transfer.readthedocs.io/en/stable/user_guide/run_ja.html#downloader-token-ja",
                }[language],
            ).run()
            if downloader_token is None:
                raise KeyboardInterrupt
            client = CommonNoLocalVolumeClient(token=downloader_token, ssl=ssl_context)
            client.test_connection()
            client.test_downloader()

        if not skip_download:
            skip_upload = button_selection(
                title="Upload Selection",
                text={
                    "en": "Do you want to upload slack data? When you want to downloaded data only, you can select No.",
                    "ja": "Slackデータのアップロードを行いますか？データをダウンロードしたいだけであれば，Noを選択してください．",
                }[language],
                choices_and_values=[("Yes", False), ("No", True)],
            )
        else:
            skip_upload = False

        uploader_token = None
        if not skip_upload:
            uploader_token = input_dialog(
                title="Input upload token",
                text={
                    "en": "Please inputs upload token. How to get is: \n"
                    + "https://slack-transfer.readthedocs.io/en/stable/user_guide/run.html#uploader-token",
                    "ja": "アップロード用のAPI tokenを入力してください．\n詳細: "
                    + "https://slack-transfer.readthedocs.io/en/stable/user_guide/run_ja.html#uploader-token-ja",
                }[language],
            ).run()
            if uploader_token is None:
                raise KeyboardInterrupt
            client = CommonNoLocalVolumeClient(token=uploader_token, ssl=ssl_context)
            client.test_connection()
            client.test_uploader()

        if not skip_download:
            confirmation(
                title="Bot invitation",
                text={
                    "en": "Please invite bot to private channels if needed. After finishing, please select OK. How to do is: \n"
                    + "https://slack-transfer.readthedocs.io/en/stable/user_guide/run.html#invite-private",
                    "ja": "必要に応じて，Botをprivate channelに招待してください．完了したら，OKを押してください．\n詳細: "
                    + "https://slack-transfer.readthedocs.io/en/stable/user_guide/run_ja.html#invite-private-ja",
                }[language],
            )

        target_all_channels = button_selection(
            title="Channel Selection",
            text={
                "en": "Do you want to process all possible channels?",
                "ja": "可能な限りすべてのチャンネルを処理の対象にしますか？",
            }[language],
            choices_and_values=[("Yes", True), ("No", False)],
        )

        channel_names = None
        if not target_all_channels:
            channels: List[Dict]
            if not skip_download:
                client = CommonNoLocalVolumeClient(
                    token=downloader_token, ssl=ssl_context
                )
                channels = client.get_channels_list()
            else:
                channels = json.load(
                    open(os.path.join(local_data_dir, "channels.json"), mode="r")
                )
            channel_names = multiple_selection(
                title="Channel selection",
                text={
                    "en": "Please select targeting channels．(Private channels listed here but bot doesn't joined will be skiped.)",
                    "ja": "処理の対象にするチャンネルを選択してください．"
                    + "なお，ここに表示されているprivate channelでも，botを招待する作業を完了していないチャンネルはスキップされます．",
                }[language],
                choices_and_values=[
                    (channel["name"], channel["name"]) for channel in channels
                ],
            )

        name_mappings = None
        override = False
        if not skip_upload:
            needs_name_mappings = button_selection(
                title="Channel name mappings",
                text={
                    "en": "Do you need to name mappings of channels when uploading data to the destination WS? "
                    + "\nDetails: https://slack-transfer.readthedocs.io/en/stable/user_guide/run.html#channel-mappings",
                    "ja": "アップロード時にチャンネル名のマッピングが必要ですか？"
                    + "\n詳細: https://slack-transfer.readthedocs.io/en/stable/user_guide/run_ja.html#channel-mappings-ja",
                }[language],
                choices_and_values=[("Yes", True), ("No", False)],
            )
            if needs_name_mappings:
                _name_mappings = input_dialog(
                    title="Channel name mappings",
                    text={
                        "en": "Please inputs name mappings. Format is comma-separated colon-concatenate dictionary: "
                        + "old_channel_name1:new_channel_name1,old_channel_name2:new_channel_name2",
                        "ja": "チャンネル名のマッピングを入力してください．カンマ区切りで，コロンでマッピング前後のチャンネル名をつないでください．"
                        + "例) old_channel_name1:new_channel_name1,old_channel_name2:new_channel_name2",
                    }[language],
                ).run()
                name_mappings = dict(
                    [
                        (dict_input.split(":")[0], dict_input.split(":")[1])
                        for dict_input in _name_mappings.split(",")
                    ]
                )

            override = button_selection(
                title="Channel override",
                text={
                    "en": "Do you want to override (add more messages) in existing channels on uploading WS?",
                    "ja": "アップロード側のWSに同じ名前のチャンネルが存在する場合に，上書き(メッセージの追加)をしますか？",
                }[language],
                choices_and_values=[("Yes", True), ("No", False)],
            )

        skip_bookmarks = button_selection(
            title="Bookmark skip",
            text={
                "en": "Do you want to skip bookmark migration? Usually select No.",
                "ja": "Bookmarkのコピーをスキップしますか？通常はNoを選択してください．",
            }[language],
            choices_and_values=[("No", False), ("Yes", True)],
        )

        print_str = (
            f"All settings:\ndata directory: {local_data_dir}\nDownload: {not skip_download}\nUpload: {not skip_upload}\n"
            + f"Downloader token: {downloader_token}\nUploader token: {uploader_token}\n"
            + f"Channels: {channel_names if channel_names else 'Not specified'}\n"
            + f"Name mappings: {name_mappings}\nOverride: {override}\nSkip bookmark: {skip_bookmarks}"
        )

        final_confirmation = button_selection(
            title="Bot invitation",
            text={
                "en": f"Please check the setting: \n{print_str}",
                "ja": f"設定を確認してください: \n{print_str}",
            }[language],
            choices_and_values=[("Continue", True), ("Cancel", False)],
        )
        if not final_confirmation:
            raise KeyboardInterrupt

        if not skip_download and not skip_upload:
            downloader = DownloaderClient(
                local_data_dir=local_data_dir, token=downloader_token
            )
            downloader.download_emoji()
            confirmation(
                title="Emoji migration required",
                text={
                    "en": f"Emojis are downloaded in {local_data_dir}/emojis/ . \n"
                    + "If you need, please add those emojis to the destination WS **NOW**. Then, click OK.",
                    "ja": f"必要に応じて，{local_data_dir}/emojis/にダウンロードされた絵文字を移行先WSに登録してください．"
                    + "完了したら，OKを押してください．",
                }[language],
            )

        run(
            local_data_dir=local_data_dir,
            downloader_token=downloader_token,
            uploader_token=uploader_token,
            override=override,
            skip_download=skip_download,
            skip_upload=skip_upload,
            channel_names=channel_names,
            name_mappings=name_mappings,
            skip_bookmarks=skip_bookmarks,
            ssl=ssl_context,
        )
    except Exception as e:
        import traceback

        print(e)
        print(traceback.format_exc())

    print(
        "This is the end of this program. Type Enter to exit. プログラムはこれで終了です．エンターキーを押すと終了します．"
    )
    while True:
        key = input()
        if not key:
            break


if __name__ == "__main__":
    print(
        f"slack_transfer {__version__} Copyright (C) M.HIRANO\nThis program comes with ABSOLUTELY NO WARRANTY"
    )
    license_file = os.path.join(os.path.dirname(__file__), "license.json")
    if os.path.exists(license_file):
        print("\nThis program including following programs:")
        license_data = json.load(open(license_file, mode="r"))
        for porg_license in license_data:
            print(
                f"{porg_license['Name']} {porg_license['Version']} Copyright (C)  {porg_license['Author']}"
            )
    interactive()



import argparse
import glob
import os
from typing import Dict
from typing import List

from slack_transfer import DownloaderClient
from slack_transfer import UploaderClient


def set_parser_bookmark(parser: argparse.ArgumentParser) -> None:
    """See bookmark section in :doc:`../../../user_guide/cli`"""
    parser.add_argument(
        "--data_dir",
        "-d",
        type=str,
        required=True,
        help="Data directory for saving download data or loading upload data. This is required.",
    )
    parser.add_argument(
        "--downloader_token",
        "-td",
        type=str,
        required=True,
        help="Download token obtained from slack (the original workspace). Starts with `xoxb-`",
    )
    parser.add_argument(
        "--uploader_token",
        "-tu",
        type=str,
        required=True,
        help="upload token obtained from slack (the destination workspace). Starts with `xoxb-`",
    )
    parser.add_argument(
        "--channel_names",
        type=str,
        default=None,
        help="channel names you want to move bookmarks. If not set, set to all available channels. "
        + "Set by comma-separation for multiple inputs. For example, `general,random`",
    )
    parser.add_argument(
        "--name_mappings",
        type=str,
        default=None,
        help="You can set name mappings between the channel names of the original and destination workspaces. "
        + "Comma-separated dictionaries (key:value) are available. For example, `old_name1:new_name1,old_name2:new_name2`.",
    )


def main_bookmark(args: argparse.Namespace) -> None:
    """See bookmark section in :doc:`../../../user_guide/cli`"""
    if args.channel_names is not None:
        channel_names = args.channel_names.split(",")
    else:
        channel_names = None

    local_data_dir = args.data_dir
    os.makedirs(local_data_dir, exist_ok=True)
    downloader = DownloaderClient(
        local_data_dir=local_data_dir, token=args.downloader_token
    )
    channels_list: List[Dict] = downloader.download_channels_list()
    if channel_names:
        channels_list = list(
            filter(lambda x: (x["name"] in channel_names), channels_list)  # type: ignore
        )
    for i, channel in enumerate(channels_list):
        downloader.download_bookmark(
            channel_id=channel["id"], channel_name=channel["name"]
        )

    uploader = UploaderClient(
        local_data_dir=local_data_dir, token=args.uploader_token, timeout=300
    )
    name_mappings = {}
    if args.name_mappings:
        name_mappings = dict(
            [
                (dict_input.split(":")[0], dict_input.split(":")[1])
                for dict_input in args.name_mappings.split(",")
            ]
        )
    uploader.create_all_channels(
        channel_names=channel_names, name_mappings=name_mappings
    )
    channel_files: List[str] = glob.glob(
        os.path.join(uploader.local_data_dir, "channels", "*.json")
    )
    if channel_names:
        channel_files = list(
            filter(
                lambda x: (os.path.basename(x)[:-5] in channel_names),  # type: ignore
                channel_files,
            )
        )
    uploader_channel_list = uploader.get_channels_list()
    for i, channel_file_path in enumerate(channel_files):
        old_channel_name = os.path.basename(channel_file_path).replace(".json", "")
        if old_channel_name in name_mappings:
            new_channel_name = name_mappings[old_channel_name]
        else:
            new_channel_name = old_channel_name
        print(f"{i + 1}/{len(channel_files)}: {old_channel_name} -> {new_channel_name}")
        channel_candidate_list = list(
            filter(lambda x: x["name"] == new_channel_name, uploader_channel_list)
        )
        if len(channel_candidate_list) != 1:
            raise IOError(
                f"channel not found in the destination channel: {new_channel_name}"
            )
        uploader.insert_bookmarks(
            channel_id=channel_candidate_list[0]["id"],
            old_channel_name=old_channel_name,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    set_parser_bookmark(parser=parser)
    args = parser.parse_args()
    main_bookmark(args=args)



import warnings
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.web import SlackResponse


def get_channels_list(client: WebClient) -> List[Dict]:
    """get channel list.

    Args:
        client (WebClient): client. If use this via any Client Class, self is automatically set. Thus, ignore this.

    Yields:
        List[Dict]: List of channel info. The details of channel info is listed: https://api.slack.com/types/channel
        (see the dictionary corresponding to channel key.)
    """
    channels: List[Dict] = []
    next_cursor: Optional[str] = None

    while True:
        response: SlackResponse = client.conversations_list(
            cursor=next_cursor, types="public_channel,private_channel"
        )
        if not response["ok"]:
            raise IOError("channel list cannot be fetched in downloading WS data.")
        channels.extend(response["channels"])

        if "response_metadata" in response:
            next_cursor = response["response_metadata"]["next_cursor"]
            if next_cursor == "":
                break
        else:
            break

    return channels


def get_replies(client: WebClient, channel_id: str, ts: str) -> List[Dict]:
    """get replies to the specific message.

    Args:
        client (WebClient): client. If use this via any Client Class, self is automatically set. Thus, ignore this.
        channel_id (str): channel id (usually 9-digits string)
        ts (str): ts of the specific message. ts is a unix timestamp in the format of float. But, here, string for it is accepted.


    Yields:
        List[Dict]: List of the reply messages. The details of message is listed: https://api.slack.com/reference/messaging/payload
    """
    messages: List[Dict] = []
    next_cursor: Optional[str] = None
    while True:
        response = client.conversations_replies(
            channel=channel_id, ts=ts, cursor=next_cursor
        )
        if not response["ok"]:
            raise IOError(
                f"replies cannot be fetched in downloading WS data. (channel_id: {channel_id}, ts: {ts})"
            )
        messages.extend(response["messages"])

        if "response_metadata" in response:
            next_cursor = response["response_metadata"]["next_cursor"]
            if next_cursor == "":
                break
        else:
            break
    return messages


def get_file_volumes(
    client: WebClient,
    channel_ids: Optional[Union[str, List[str]]] = None,
    auto_join: bool = False,
) -> float:
    """Calculate the total file volumes (bytes)

    Args:
        client (WebClient): client. If use this via any Client Class, self is automatically set. Thus, ignore this.
        channel_ids (Optional; str or List[str]; default=None): channel id list for calculation.
            If not set, calculate all accessible channels.
        auto_join (bool; default=False): when bot is not in channel, automatically join if possible (public channel).
            This feature is disabled for :class:`slack_transfer._base.CommonDryRunClient`

    Yields:
        float: total file volumes (bytes)
    """
    channels = get_channels_list(client=client)
    channel_id_name_dict = dict([(x["id"], x["name"]) for x in channels])
    if channel_ids is None:
        channel_ids = list(map(lambda x: x["id"], channels))
    elif isinstance(channel_ids, str):
        channel_ids = [channel_ids]
    else:
        channel_ids = channel_ids

    volume_dict: Dict[str, int] = {}

    for channel_id in channel_ids:
        next_cursor: Optional[str] = None
        while True:
            try:
                response: SlackResponse = client.files_list(
                    channel=channel_id, cursor=next_cursor
                )
            except SlackApiError as e:
                if e.response["error"] == "not_in_channel":
                    if auto_join:
                        try:
                            client.conversations_join(channel=channel_id)
                            continue
                        except:
                            pass
                    warnings.warn(
                        f"slack bot is not in `{channel_id_name_dict[channel_id] if channel_id in channel_id_name_dict else channel_id}`. Skip this."
                    )
                    break
                else:
                    raise e
            if not response["ok"]:
                raise IOError("channel list cannot be fetched.")
            volume_dict.update(
                dict(map(lambda x: (x["id"], x["size"]), response["files"]))
            )

            if "response_metadata" in response:
                next_cursor = response["response_metadata"]["next_cursor"]
                if next_cursor == "":
                    break
            else:
                break
    return sum(volume_dict.values())


def test_connection(client: WebClient) -> None:
    """test connections to slack.

    This test including common scope among all client.
    The token set to the client use this method requires the following scope:
     - channels:history
     - channels:join
     - channels:read
     - files:read
     - groups:history
     - groups:read

     If one of the scope is missing, this method rise an error.
     Those tests are performed step by step. It means that missing scope appears not at once.

    Args:
        client (WebClient): client. If use this via any Client Class, self is automatically set. Thus, ignore this.
    """
    response = client.auth_test()
    if not response["ok"]:
        raise IOError("slack token is invalid.")
    print(f"Successfully access to `{response['team']}`")

    # channels:read,groups:read
    try:
        response = client.conversations_list(types="public_channel,private_channel")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise AttributeError("missing scope: channels:read,groups:read")
        else:
            raise e

    channel_cand_public = list(filter(lambda x: x["is_general"], response.data["channels"]))  # type: ignore
    if len(channel_cand_public) == 0:
        raise ValueError("general type channel doesn't exist.")
    if len(channel_cand_public) != 1:
        raise AssertionError

    # channels:join
    try:
        client.conversations_join(channel=channel_cand_public[0]["id"])
    except SlackApiError as e:
        if e.response["error"] != "method_not_supported_for_channel_type":
            raise e
    # channels:history for public
    try:
        client.conversations_history(channel=channel_cand_public[0]["id"], ts="1")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise AttributeError("missing scope: channels:history")
        else:
            raise e
    # files:read
    client.files_list()

    channel_cand_private = list(filter(lambda x: x["is_private"], response.data["channels"]))  # type: ignore
    if len(channel_cand_private) == 0:
        warnings.warn("No private channel is tested")
        return
    # groups:history for private
    try:
        client.conversations_history(channel=channel_cand_private[0]["id"], ts="1")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise AttributeError("missing scope: groups:history")
        else:
            raise e


def test_downloader(client: WebClient) -> None:
    """test connections to slack as downloader.

    This test including common scope for DownloaderClient.
    The token set to the client use this method requires the following scope:
     - bookmarks:read
     - emoji:read
     - users:read

     If one of the scope is missing, this method rise an error.
     Those tests are performed step by step. It means that missing scope appears not at once.

    Args:
        client (WebClient): client. If use this via any Client Class, self is automatically set. Thus, ignore this.
    """
    client.users_list()

    # channels:read,groups:read
    response = client.conversations_list(types="public_channel,private_channel")
    channel_cand_public = list(filter(lambda x: x["is_general"], response.data["channels"]))  # type: ignore
    if len(channel_cand_public) == 0:
        raise ValueError("general type channel doesn't exist.")
    if len(channel_cand_public) != 1:
        raise AssertionError
    # bookmarks:read
    client.bookmarks_list(channel_id=channel_cand_public[0]["id"])
    # emoji:read
    client.emoji_list()
    print("test for downloader is finished!")


def test_uploader(client: WebClient) -> None:
    """test connections to slack as uploader.

    This test including common scope for UploaderClient.
    The token set to the client use this method requires the following scope:
     - channels:manage
     - files:write
     - chat:write
     - pins:write
     - bookmarks:write
     - reaction:write

     If one of the scope is missing, this method rise an error.
     Those tests are performed step by step. It means that missing scope appears not at once.

    Args:
        client (WebClient): client. If use this via any Client Class, self is automatically set. Thus, ignore this.
    """
    # channels:manage
    # conversations_setTopic and conversations_setPurpose is also the same scope
    try:
        client.conversations_create(name=":::", is_private=False)
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise AttributeError("missing scope: channels:manage")
        elif e.response["error"] == "invalid_name_specials":
            pass
        else:
            raise e
    # try:
    #     client.conversations_create(name=":::", is_private=True)
    # except SlackApiError as e:
    #     if e.response["error"] == "missing_scope":
    #         raise AttributeError(f"missing scope: groups:write")
    #     elif e.response["error"] == "invalid_name_specials":
    #         pass
    #     else:
    #         raise e

    # files:write
    try:
        client.files_upload(content="aaa", channels=":::::")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise e

    # files_list is tested in test_connection

    # chat:write
    try:
        client.chat_postMessage(channel=":::::", text="test")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise AttributeError("missing scope: chat:write")

    # pins:write
    try:
        client.pins_add(channel=":::::", timestamp="1")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise e

    # bookmarks:write
    try:
        client.bookmarks_add(channel_id=":::::", title="test", type="link")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise e

    # reaction:write
    try:
        client.reactions_add(channel=":::::", name="test", timestamp="1")
    except SlackApiError as e:
        if e.response["error"] == "missing_scope":
            raise e

    print("test for uploader is finished!")



import json
import mimetypes
import os
import time
import warnings
from typing import Dict
from typing import List
from typing import Optional

import requests
import tqdm.std
from slack_sdk.errors import SlackApiError
from slack_sdk.web import SlackResponse

from .._base import DownloaderClientABC
from .common import get_channels_list
from .common import get_replies


def download_channels_list(client: DownloaderClientABC) -> List[Dict]:
    """download and save channel list.

    Args:
        client (DownloaderClientABC): downloader client. If use this via any DownloaderClient Class, self is automatically set.
            Thus, ignore this.

    Yields:
        List[Dict]: List of channel info. The details of channel info is listed: https://api.slack.com/types/channel
        (see the dictionary corresponding to channel key.)
    """
    channels: List[Dict] = get_channels_list(client=client)

    json.dump(
        channels,
        open(
            os.path.join(client.local_data_dir, "channels.json"),
            mode="w",
            encoding="utf-8",
        ),
        indent=4,
    )
    return channels


def download_file(
    client: DownloaderClientABC, file_id: str, file_name: str, url_private: str
) -> None:
    """download and save a specified file.

    The file will be saved as file_id--file_name. Most parameters are specified by slack and **this method isn't used by users.**

    Args:
        client (DownloaderClientABC): downloader client. If use this via any DownloaderClient Class, self is automatically set.
            Thus, ignore this.
        file_id (str): file id on slack (for saving and used for identify the correct file by uploader).
        file_name (str): fila name for saving.
        url_private (str): private url specified by slack
    """
    res = requests.get(
        url=url_private,
        allow_redirects=True,
        headers={"Authorization": f"Bearer {client.token}"},
        stream=True,
    )
    file_name = f"{file_id}--{file_name}"
    if res.status_code != 200:
        warnings.warn(f"failed to download: {url_private} as {file_name}")
        return None
    file_path = os.path.join(client.local_data_dir, "files", file_name)
    with open(file_path, mode="wb") as f:
        f.write(res.content)


def download_channel_history(
    client: DownloaderClientABC,
    channel_id: str,
    channel_name: str,
    latest: Optional[str] = None,
    ts_progress_bar: Optional[tqdm.tqdm] = None,
    ts_now: Optional[int] = None,
    auto_join: bool = True,
) -> None:
    """download and save a series of channel histories on a channel.

    The file will be saved as file_id--file_name. Most parameters are specified by slack and **this method isn't used by users.**

    Args:
        client (DownloaderClientABC): downloader client. If use this via any DownloaderClient Class, self is automatically set.
            Thus, ignore this.
        channel_id (str): channel id (usually 9-digits string)
        channel_name (str): channel name. Please set it correctly. Usually this can be obtained using channel_id.
            But, because it is usually redundant, please set from outside.
        latest (str; Optional; default=None): the latest time of the time range you want to fetch.
            Set as float-like string representing unix timetamp including floating point.
        ts_progress_bar (tqdm.tqdm: Optional; default=None): progress bar instance.
            Progress bar will be updated using the diff between ts_now and the oldest fetched history.
        ts_now (int; Optional; default=None): anchor point to calculate the progress.
            It is required when ts_progress_bar is set.
        auto_join (bool; Optional; default=True): when bot is not in channel, automatically join if possible (public channel).
    """
    # ToDo: 1 channel内でAPI limit来た場合の挙動
    if ts_progress_bar:
        if ts_now is None:
            ts_now = int(time.time())

    def _download_files_in_message(message: Dict) -> None:
        if "files" in message:
            for file in message["files"]:
                if "url_private" in file:
                    url_private: str = file["url_private"]
                    file_id: str = file["id"]
                    file_name: str = file["name"]
                    download_file(
                        client=client,
                        file_id=file_id,
                        file_name=file_name,
                        url_private=url_private,
                    )

    messages = []
    next_cursor: Optional[str] = None
    while True:
        try:
            response = client.conversations_history(
                channel=channel_id,
                latest=latest,
                cursor=next_cursor,
                include_all_metadata=True,
            )
        except SlackApiError as e:
            if e.response["error"] == "not_in_channel":
                if auto_join:
                    try:
                        client.conversations_join(channel=channel_id)
                        continue
                    except:
                        pass
                warnings.warn(f"slack bot is not in `{channel_name}`. Skip this.")
                return None
            else:
                raise e
        if not response["ok"]:
            raise IOError(
                f"channel history cannot be fetched in downloading WS data. (channel_id: {channel_id}, channel_name: {channel_name}, latest: {latest})"
            )
        for _message in response["messages"]:
            _download_files_in_message(message=_message)
            if "reply_count" in _message and _message["reply_count"] > 0:
                ts: str = _message["ts"]
                replies = get_replies(client=client, channel_id=channel_id, ts=ts)
                messages.extend(replies)
                for reply in replies:
                    _download_files_in_message(message=reply)
            else:
                messages.append(_message)

        if ts_progress_bar and len(messages) > 0:
            if ts_now is None:
                raise AssertionError
            oldest_ts = messages[-1]["ts"]
            progress_ts = int(ts_now - float(oldest_ts))
            update_p = progress_ts - ts_progress_bar.n
            ts_progress_bar.update(update_p)

        if "response_metadata" in response:
            next_cursor = response["response_metadata"]["next_cursor"]
            latest = None
            if next_cursor == "":
                break
        else:
            break
    messages = list(
        {
            message["client_msg_id"]: message
            for message in [
                message for message in messages if "client_msg_id" in message
            ]
        }.values()
    ) + [message for message in messages if "client_msg_id" not in message]
    messages.sort(key=lambda x: x["ts"], reverse=False)
    json.dump(
        messages,
        open(
            os.path.join(client.local_data_dir, "channels", f"{channel_name}.json"),
            mode="w",
            encoding="utf-8",
        ),
        indent=4,
    )


def download_members_list(client: DownloaderClientABC) -> List[Dict]:
    """download and save a member list.

    Args:
        client (DownloaderClientABC): downloader client. If use this via any DownloaderClient Class, self is automatically set.
            Thus, ignore this.
    """
    members: List[Dict] = []
    next_cursor: Optional[str] = None

    while True:
        response: SlackResponse = client.users_list(cursor=next_cursor)
        if not response["ok"]:
            raise IOError("user list cannot be fetched in downloading WS data.")
        members.extend(response["members"])

        if "response_metadata" in response:
            next_cursor = response["response_metadata"]["next_cursor"]
            if next_cursor == "":
                break
        else:
            break

    json.dump(
        members,
        open(
            os.path.join(client.local_data_dir, "members.json"),
            mode="w",
            encoding="utf-8",
        ),
        indent=4,
    )
    return members


def download_bookmark(
    client: DownloaderClientABC,
    channel_id: str,
    channel_name: str,
    auto_join: bool = True,
) -> None:
    """download and save a series of bookmarks on a channel.

    Args:
        client (DownloaderClientABC): downloader client. If use this via any DownloaderClient Class, self is automatically set.
            Thus, ignore this.
        channel_id (str): channel id (usually 9-digits string)
        channel_name (str): channel name. Please set it correctly. Usually this can be obtained using channel_id.
            But, because it is usually redundant, please set from outside.
        auto_join (bool; Optional; default=True): when bot is not in channel, automatically join if possible (public channel).
    """
    for _ in range(3):
        try:
            response = client.bookmarks_list(channel_id=channel_id)
        except SlackApiError as e:
            if e.response["error"] == "not_in_channel":
                if auto_join:
                    try:
                        client.conversations_join(channel=channel_id)
                        continue
                    except:
                        pass
                warnings.warn(f"slack bot is not in `{channel_name}`. Skip this.")
                return None
            else:
                raise e
        if not response["ok"]:
            raise IOError(
                f"bookmarks cannot be fetched in downloading WS data. (channel_id: {channel_id}, channel_name: {channel_name})"
            )
        if "bookmarks" in response:
            json.dump(
                response["bookmarks"],
                open(
                    os.path.join(
                        client.local_data_dir, "bookmarks", f"{channel_name}.json"
                    ),
                    mode="w",
                    encoding="utf-8",
                ),
                indent=4,
            )
        break


def download_emoji(client: DownloaderClientABC) -> None:
    """download and save a moji list and emoji files.

    Args:
        client (DownloaderClientABC): downloader client. If use this via any DownloaderClient Class, self is automatically set.
            Thus, ignore this.
    """
    response = client.emoji_list()
    if "emoji" in response:
        json.dump(
            response["emoji"],
            open(
                os.path.join(client.local_data_dir, "emoji.json"),
                mode="w",
                encoding="utf-8",
            ),
            indent=4,
        )
        for key, url in response["emoji"].items():
            if not url.startswith("alias:"):
                res = requests.get(
                    url=url,
                    allow_redirects=True,
                    headers={"Authorization": f"Bearer {client.token}"},
                    stream=True,
                )
                extension = mimetypes.guess_extension(
                    res.headers.get("content-type", default="image/png")
                )
                file_name = f"{key}{extension}"
                if res.status_code != 200:
                    warnings.warn(f"failed to download: {url} as {file_name}")
                    return None
                file_path = os.path.join(client.local_data_dir, "emojis", file_name)
                with open(file_path, mode="wb") as f:
                    f.write(res.content)



import argparse

from .._base import CommonNoLocalVolumeClient


def set_parser_token_test(parser: argparse.ArgumentParser) -> None:
    """See token_test section in :doc:`../../../user_guide/cli`"""
    parser.add_argument(
        "--token",
        type=str,
        required=True,
        help="Token obtained from slack. Starts with `xoxb-`",
    )
    parser.add_argument(
        "--as_downloader", action="store_true", help="for checking as downloader"
    )
    parser.add_argument(
        "--as_uploader", action="store_true", help="for checking as uploader"
    )


def main_token_test(args: argparse.Namespace) -> None:
    """See token_test section in :doc:`../../../user_guide/cli`"""
    client = CommonNoLocalVolumeClient(token=args.token)
    client.test_connection()

    if args.as_downloader:
        client.test_downloader()

    if args.as_uploader:
        client.test_uploader()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    set_parser_token_test(parser=parser)
    args = parser.parse_args()
    main_token_test(args=args)
